@@ -148,10 +148,12 @@ class GeneralSection (Section):
 
 
     def on_systray_check_toggled(self, widget):
-        """Systray option controls sensitivity of skip taskbar"""
+        """Systray option controls sensitivity of sub-options"""
         self.xml.get_widget("skip_taskbar_check").set_sensitive(
             widget.get_active())
-        
+        self.xml.get_widget("minimize_on_start_check").set_sensitive(
+            widget.get_active())
+
     
     def on_set_default_notebook_button_clicked(self, widget):
 
@@ -167,59 +169,94 @@ class GeneralSection (Section):
             self.notebook = win.get_notebook()
 
         # populate default notebook        
-        if app.pref.use_last_notebook:
+        if app.pref.get("use_last_notebook", default=True):
             self.xml.get_widget("last_notebook_radio").set_active(True)
-        elif app.pref.default_notebook == "":
+        elif app.pref.get("default_notebooks", default=[]) == []:
             self.xml.get_widget("no_default_notebook_radio").set_active(True)
         else:
             self.xml.get_widget("default_notebook_radio").set_active(True)
             self.xml.get_widget("default_notebook_entry").\
-                set_text(app.pref.default_notebook)
-
+                set_text(
+                (app.pref.get("default_notebooks", default=[]) + [""])[0]
+                )
 
 
         # populate autosave
         self.xml.get_widget("autosave_check").set_active(
-            app.pref.autosave)
+            app.pref.get("autosave"))
         self.xml.get_widget("autosave_entry").set_text(
-            str(int(app.pref.autosave_time / 1000)))
+            str(int(app.pref.get("autosave_time") / 1000)))
 
         self.xml.get_widget("autosave_entry").set_sensitive(
-            app.pref.autosave)
+            app.pref.get("autosave"))
         self.xml.get_widget("autosave_label").set_sensitive(
-            app.pref.autosave)
+            app.pref.get("autosave"))
 
 
         # use systray icon
-        self.xml.get_widget("systray_check").set_active(app.pref.use_systray)
-        self.xml.get_widget("skip_taskbar_check").set_active(app.pref.skip_taskbar)
-        self.xml.get_widget("skip_taskbar_check").set_sensitive(app.pref.use_systray)
+        self.xml.get_widget("systray_check").set_active(
+            app.pref.get("window", "use_systray"))
+        self.xml.get_widget("skip_taskbar_check").set_active(
+            app.pref.get("window", "skip_taskbar"))
+        self.xml.get_widget("skip_taskbar_check").set_sensitive(
+            app.pref.get("window", "use_systray"))
+
+        self.xml.get_widget("minimize_on_start_check").set_active(
+            app.pref.get("window", "minimize_on_start"))
+        self.xml.get_widget("minimize_on_start_check").set_sensitive(
+            app.pref.get("window", "use_systray"))
+
+        self.xml.get_widget("window_keep_above_check").set_active(
+            app.pref.get("window", "keep_above"))
+            
+        
+        # set window 'always on top'
+        self.xml.get_widget("window_stick_check").set_active(
+            app.pref.get("window", "stick"))
 
 
     def save_options(self, app):
         if self.xml.get_widget("last_notebook_radio").get_active():
-            app.pref.use_last_notebook = True
+            app.pref.set("use_last_notebook", True)
         elif self.xml.get_widget("default_notebook_radio").get_active():
-            app.pref.use_last_notebook = False
-            app.pref.default_notebook = \
-                unicode_gtk(self.xml.get_widget("default_notebook_entry").get_text())
+            app.pref.set("use_last_notebook", False)
+            app.pref.set("default_notebooks", 
+                         [unicode_gtk(
+                        self.xml.get_widget(
+                            "default_notebook_entry").get_text())])
         else:
-            app.pref.use_last_notebook = False
-            app.pref.default_notebook = ""
+            app.pref.set("use_last_notebook", False)
+            app.pref.set("default_notebooks", [])
 
 
         # save autosave
-        app.pref.autosave = \
-            self.xml.get_widget("autosave_check").get_active()
+        app.pref.set("autosave", 
+            self.xml.get_widget("autosave_check").get_active())
         try:
-            app.pref.autosave_time = \
-                int(self.xml.get_widget("autosave_entry").get_text()) * 1000
+            app.pref.set("autosave_time",
+                int(self.xml.get_widget("autosave_entry").get_text()) * 1000)
         except:
             pass
 
         # use systray icon
-        app.pref.use_systray = self.xml.get_widget("systray_check").get_active()
-        app.pref.skip_taskbar = self.xml.get_widget("skip_taskbar_check").get_active()
+        app.pref.set("window", "use_systray", 
+                     self.xml.get_widget("systray_check").get_active())
+        app.pref.set("window", "skip_taskbar", 
+                     self.xml.get_widget("skip_taskbar_check").get_active())
+
+        app.pref.set(
+            "window", "minimize_on_start", 
+            self.xml.get_widget("minimize_on_start_check").get_active())
+ 
+        # window 'always above'
+        app.pref.set(
+            "window", "keep_above", 
+            self.xml.get_widget("window_keep_above_check").get_active())
+
+        # window 'stick to all desktops'
+        app.pref.set(
+            "window", "stick",
+            self.xml.get_widget("window_stick_check").get_active())
 
 
 class LookAndFeelSection (Section):
@@ -246,6 +283,22 @@ class LookAndFeelSection (Section):
             _("use GTK stock icons in toolbar"))
         self.use_minitoolbar = add_check(
             _("use minimal toolbar"))
+
+
+        # app font size
+        font_size = 10
+        h = gtk.HBox(False, 5); h.show()
+        l = gtk.Label(_("Application Font Size:")); l.show()
+        h.pack_start(l, False, False, 0)
+        self.app_font_size = gtk.SpinButton(
+          gtk.Adjustment(value=font_size, lower=2, upper=500, step_incr=1))
+        self.app_font_size.set_value(font_size)
+        #font_size_button.set_editable(False)
+        self.app_font_size.show()
+        h.pack_start(self.app_font_size, False, False, 0)
+        v.pack_start(h, False, False, 0)
+
+
         
         # view mode combo
         h = gtk.HBox(False, 5); h.show()
@@ -262,12 +315,14 @@ class LookAndFeelSection (Section):
 
     def load_options(self, app):
 
-        self.treeview_lines_check.set_active(app.pref.treeview_lines)
-        self.listview_rules_check.set_active(app.pref.listview_rules)
-        self.use_stock_icons_check.set_active(app.pref.use_stock_icons)
-        self.use_minitoolbar.set_active(app.pref.use_minitoolbar)
+        l = app.pref.get("look_and_feel")
+        self.treeview_lines_check.set_active(l.get("treeview_lines"))
+        self.listview_rules_check.set_active(l.get("listview_rules"))
+        self.use_stock_icons_check.set_active(l.get("use_stock_icons"))
+        self.use_minitoolbar.set_active(l.get("use_minitoolbar"))
+        self.app_font_size.set_value(l.get("app_font_size"))
 
-        if app.pref.view_mode == "horizontal":
+        if app.pref.get("viewers", "three_pane_viewer", "view_mode", default="") == "horizontal":
             self.listview_layout.set_active(1)
         else:
             self.listview_layout.set_active(0)
@@ -276,13 +331,16 @@ class LookAndFeelSection (Section):
 
     def save_options(self, app):
         
-        app.pref.treeview_lines = self.treeview_lines_check.get_active()
-        app.pref.listview_rules = self.listview_rules_check.get_active()
-        app.pref.use_stock_icons = self.use_stock_icons_check.get_active()
-        app.pref.use_minitoolbar = self.use_minitoolbar.get_active()
+        l = app.pref.get("look_and_feel")
+        l["treeview_lines"] = self.treeview_lines_check.get_active()
+        l["listview_rules"] = self.listview_rules_check.get_active()
+        l["use_stock_icons"] = self.use_stock_icons_check.get_active()
+        l["use_minitoolbar"] = self.use_minitoolbar.get_active()
+        l["app_font_size"] = self.app_font_size.get_value()
 
-        app.pref.view_mode = ["vertical", "horizontal"][
-            self.listview_layout.get_active()]
+        app.pref.set("viewers", "three_pane_viewer", "view_mode", 
+                     ["vertical", "horizontal"][
+                self.listview_layout.get_active()])
  
 
 class LanguageSection (Section):
@@ -315,12 +373,14 @@ class LanguageSection (Section):
 
     def load_options(self, app):
 
+        lang = app.pref.get("language", default="")
+
         # set default
-        if app.pref.language == "":
+        if lang == "":
             self.language_box.set_active(0)
         else:
             for i, row in enumerate(self.language_box.get_model()):
-                if app.pref.language == row[0]:
+                if lang == row[0]:
                     self.language_box.set_active(i)
                     break
 
@@ -328,10 +388,10 @@ class LanguageSection (Section):
     def save_options(self, app):
         
         if self.language_box.get_active() > 0:
-            app.pref.language = lang = self.language_box.get_active_text()
+            app.pref.set("language", self.language_box.get_active_text())
         else:
             # set default
-            app.pref.language = ""
+            app.pref.set("language", "")
 
         # XXX: may be I should not change translation during execution
         #if app.pref.language != keepnote.trans.get_lang():
@@ -348,13 +408,13 @@ class HelperAppsSection (Section):
         self.entries = {}
         w = self.get_default_widget()
         
-        self.table = gtk.Table(len(app.pref.external_apps), 2)
+        self.table = gtk.Table(max(len(list(app.iter_external_apps())), 1), 2)
         self.table.show()
         w.add(self.table)
 
         # set icon
         try:
-            self.icon = keepnote.gui.get_pixbuf(get_icon_filename(gtk.STOCK_EXECUTE))
+            self.icon = keepnote.gui.get_pixbuf(get_icon_filename(gtk.STOCK_EXECUTE), size=(15, 15))
         except:
             pass
 
@@ -364,9 +424,9 @@ class HelperAppsSection (Section):
 
         # clear table, resize
         self.table.foreach(lambda x: self.table.remove(x))
-        self.table.resize(len(app.pref.external_apps), 2)
+        self.table.resize(len(list(app.iter_external_apps())), 2)
 
-        for i, app in enumerate(app.pref.external_apps):
+        for i, app in enumerate(app.iter_external_apps()):
             key = app.key
             app_title = app.title
             prog = app.prog
@@ -403,8 +463,8 @@ class HelperAppsSection (Section):
             button.show()
             button.connect("clicked", button_clicked(key, app_title, prog))
             self.table.attach(button, 2, 3, i, i+1,
-                         xoptions=0, yoptions=0,
-                         xpadding=2, ypadding=2)
+                              xoptions=0, yoptions=0,
+                              xpadding=2, ypadding=2)
     
 
     def save_options(self, app):
@@ -412,12 +472,14 @@ class HelperAppsSection (Section):
         # TODO: use a public interface
 
         # save external app options
-        for key, entry in self.entries.iteritems():
-            if key in app.pref._external_apps_lookup:
-                app.pref._external_apps_lookup[key].prog = unicode_gtk(
-                    entry.get_text())
-
+        apps = app.pref.get("external_apps", default=[])
 
+        for app in apps:
+            key = app.get("key", None)
+            if key:
+                entry = self.entries.get(key, None)
+                if entry:
+                    app["prog"] = unicode_gtk(entry.get_text())
 
 
 class DatesSection (Section):
@@ -434,13 +496,13 @@ class DatesSection (Section):
     def load_options(self, app):
         for name in ["same_day", "same_month", "same_year", "diff_year"]:
             self.date_xml.get_widget("date_%s_entry" % name).\
-                set_text(app.pref.timestamp_formats[name])
+                set_text(app.pref.get("timestamp_formats", name))
 
     def save_options(self, app):
         # save date formatting
         for name in ["same_day", "same_month", "same_year", "diff_year"]:
-            app.pref.timestamp_formats[name] = unicode_gtk(
-                self.date_xml.get_widget("date_%s_entry" % name).get_text())
+            app.pref.set("timestamp_formats", name, unicode_gtk(
+                self.date_xml.get_widget("date_%s_entry" % name).get_text()))
         
 
 class AllNoteBooksSection (Section):
@@ -493,12 +555,15 @@ class NoteBookSection (Section):
     def load_options(self, app):
         
         if self.notebook is not None:
-            font = self.notebook.pref.default_font
+            font = self.notebook.pref.get("default_font", 
+                                          default=keepnote.gui.DEFAULT_FONT)
             family, mods, size = keepnote.gui.richtext.parse_font(font)
             self.notebook_font_family.set_family(family)
             self.notebook_font_size.set_value(size)
 
-            self.notebook_index_dir.set_text(self.notebook.pref.index_dir)
+            self.notebook_index_dir.set_text(
+                self.notebook.pref.get("index_dir", 
+                                       default=u"", type=basestring))
 
 
     def save_options(self, app):
@@ -507,12 +572,12 @@ class NoteBookSection (Section):
             pref = self.notebook.pref
 
             # save notebook font        
-            pref.default_font = "%s %d" % (
-                self.notebook_font_family.get_family(),
-                self.notebook_font_size.get_value())
+            pref.set("default_font", "%s %d" % (
+                    self.notebook_font_family.get_family(),
+                    self.notebook_font_size.get_value()))
 
             # alternative index directory
-            pref.index_dir = self.notebook_index_dir.get_text()
+            pref.set("index_dir",  self.notebook_index_dir.get_text())
 
         
 
@@ -564,7 +629,7 @@ class ExtensionsSection (Section):
 
         # set icon
         try:
-            self.icon = keepnote.gui.get_pixbuf(get_icon_filename(gtk.STOCK_ADD))
+            self.icon = keepnote.gui.get_pixbuf(get_icon_filename(gtk.STOCK_ADD), size=(15, 15))
         except:
             pass
 
@@ -576,10 +641,10 @@ class ExtensionsSection (Section):
         self.extlist.foreach(self.extlist.remove)
         
         def callback(ext):
-            return lambda w: self._on_uninstall(ext)
+            return lambda w: self._on_uninstall(ext.key)
 
         # populate extension list
-        exts = list(app.iter_extensions())
+        exts = list(app.get_imported_extensions())
         d = {"user": 0, "system": 1}
         exts.sort(key=lambda e: (d.get(e.type, 10), e.name))
         for ext in exts:
@@ -598,12 +663,19 @@ class ExtensionsSection (Section):
 
     def save_options(self, app):
         
+        app.pref.set(
+            "extension_info", "disabled", 
+            [widget.ext.key for widget in self.extlist
+             if not widget.enabled])
+
+        # enable/disable extensions
         for widget in self.extlist:
-            try:
-                widget.ext.enable(widget.enabled)
-            except Exception, e:
-                keepnote.log_error(e)
-            widget.update()
+            if widget.enabled != widget.ext.is_enabled():
+                try:
+                    widget.ext.enable(widget.enabled)
+                except:
+                    keepnote.log_error()
+                
 
 
     def _on_uninstall(self, ext):
@@ -692,7 +764,7 @@ class ExtensionWidget (gtk.EventBox):
         v.pack_start(h, True, True, 0)
 
         # enable button
-        self.enable_check = gtk.CheckButton("Enabled")
+        self.enable_check = gtk.CheckButton(_("Enabled"))
         self.enable_check.set_active(self.enabled)
         self.enable_check.show()
         self.enable_check.connect(
@@ -704,7 +776,7 @@ class ExtensionWidget (gtk.EventBox):
         h.pack_start(l, False, True, 0)
 
         # uninstall button        
-        self.uninstall_button = gtk.Button("Uninstall")
+        self.uninstall_button = gtk.Button(_("Uninstall"))
         self.uninstall_button.set_relief(gtk.RELIEF_NONE)
         self.uninstall_button.set_sensitive(app.can_uninstall(ext))
         self.uninstall_button.show()
@@ -732,7 +804,6 @@ class ApplicationOptionsDialog (object):
         self.parent = None
 
         self._sections = []
-        self.tree2section = {}
         
         self.xml = gtk.glade.XML(get_resource("rc", "keepnote.glade"),
                                  "app_options_dialog", keepnote.GETTEXT_DOMAIN)
@@ -763,12 +834,11 @@ class ApplicationOptionsDialog (object):
         column.pack_start(cell_text, True)
         column.add_attribute(cell_text, 'text', 0)
         
-
         # add tabs
         self.add_default_sections()
 
     
-    def show(self, parent):
+    def show(self, parent, section=None):
         """Display application options"""
         
         self.parent = parent
@@ -786,7 +856,7 @@ class ApplicationOptionsDialog (object):
 
         # add extension options
         self.extensions_ui = []
-        for ext in self.app.iter_extensions(True):
+        for ext in self.app.get_enabled_extensions():
             if isinstance(ext, keepnote.gui.extension.Extension):
                 ext.on_add_options_ui(self)
                 self.extensions_ui.append(ext)
@@ -796,6 +866,13 @@ class ApplicationOptionsDialog (object):
 
         self.dialog.show()
 
+        if section:
+            try:
+                self.overview.set_cursor(self.get_section_path(section))
+            except:
+                pass
+
+
     def finish(self):
 
         # remove extension options
@@ -849,17 +926,24 @@ class ApplicationOptionsDialog (object):
     
     def save_options(self, app):
         """Save the options from each section"""
+
+        # let app record its preferences first
+        app.save_preferences()
         
+        # let sections record their preferences
         for section in self._sections:
             section.save_options(self.app)
-
-        # notify application preference changes
+        
+        # notify changes
+        # app and notebook will load prefs from plist
         self.app.pref.changed.notify()
-
-        # save noteboook preference changes
         for notebook in self.app.iter_notebooks():
-            notebook.write_preferences()
             notebook.notify_change(False)
+        
+        # force a app and notebook preference save
+        # save prefs to plist and to disk
+        app.save()
+
 
 
     #=====================================
@@ -868,6 +952,9 @@ class ApplicationOptionsDialog (object):
     def add_section(self, section, parent=None):
         """Add a section to the Options Dialog"""
         
+        # icon size
+        size = (15, 15)
+
         # determine parent section
         if parent is not None:
             path = self.get_section_path(parent)
@@ -885,7 +972,7 @@ class ApplicationOptionsDialog (object):
             icon = icon="note.png"
         
         if isinstance(icon, basestring):
-            pixbuf = keepnote.gui.get_resource_pixbuf(icon)
+            pixbuf = keepnote.gui.get_resource_pixbuf(icon, size=size)
         else:
             pixbuf = icon
 
@@ -893,29 +980,24 @@ class ApplicationOptionsDialog (object):
         it = self.overview_store.append(it, [section.label, section, pixbuf])
         path = self.overview_store.get_path(it)
         self.overview.expand_to_path(path)
-        self.tree2section[path] = section
 
         return section
 
 
     def remove_section(self, key):
         
-        # TODO: may need to update tree2section, when other pages slide in position
-
-        path = self.get_section_path(key)
-
-        if path is None:
-            return
-
         # remove from tabs
-        section = self.tree2section[path]
-        self.tabs.remove_page(self._sections.index(section))
-        del self.tree2section[path]
+        section = self.get_section(key)
+        if section:
+            self.tabs.remove_page(self._sections.index(section))
+            self._sections.remove(section)
 
         # remove from tree
-        self.overview_store.remove(self.overview_store.get_iter(path))
+        path = self.get_section_path(key)
+        if path is not None:
+            self.overview_store.remove(self.overview_store.get_iter(path))
+
 
-        self._sections.remove(section)
 
 
     def get_section(self, key):
@@ -924,6 +1006,7 @@ class ApplicationOptionsDialog (object):
         for section in self._sections:
             if section.key == key:
                 return section
+        return None
 
             
     def get_section_path(self, key):
@@ -955,7 +1038,7 @@ class ApplicationOptionsDialog (object):
         """Callback for changing topic in overview"""
         row, col = overview.get_cursor()
         if row is not None:
-            section = self.tree2section[row]
+            section = self.overview_store[row][1]
             self.tabs.set_current_page(self._sections.index(section))
 
 
