@@ -35,6 +35,7 @@ import re
 import random
 import urllib2
 import StringIO
+from itertools import chain
 
 
 # pygtk imports
@@ -52,11 +53,11 @@ except ImportError:
 
 
 # textbuffer_tools imports
-from keepnote.gui.richtext.textbuffer_tools import \
-     iter_buffer_contents
+from .textbuffer_tools import \
+     iter_buffer_contents, iter_buffer_anchors, sanitize_text
 
 # richtextbuffer imports
-from keepnote.gui.richtext.richtextbuffer import \
+from .richtextbuffer import \
      ignore_tag, \
      add_child_to_buffer, \
      RichTextBuffer, \
@@ -64,7 +65,7 @@ from keepnote.gui.richtext.richtextbuffer import \
      RichTextIndentTag
 
 # tag imports
-from keepnote.gui.richtext.richtext_tags import \
+from .richtext_tags import \
      RichTextModTag, \
      RichTextJustifyTag, \
      RichTextFamilyTag, \
@@ -78,7 +79,7 @@ from keepnote.gui.richtext.richtext_tags import \
      set_text_scale
 
 # richtext io
-from keepnote.gui.richtext.richtext_html import HtmlBuffer, HtmlError
+from .richtext_html import HtmlBuffer, HtmlError
 
 from keepnote import safefile
 
@@ -92,13 +93,16 @@ _ = keepnote.translate
 # constants
 DEFAULT_FONT = "Sans 10"
 TEXTVIEW_MARGIN = 5
-CLIPBOARD_NAME = "CLIPBOARD"
+if keepnote.get_platform() == "darwin":
+    CLIPBOARD_NAME = gdk.SELECTION_PRIMARY
+else:
+    CLIPBOARD_NAME = "CLIPBOARD"
 RICHTEXT_ID = -3    # application defined integer for the clipboard
-
+CONTEXT_MENU_ACCEL_PATH = "<main>/richtext_context_menu"
 
 # mime types
-# keepnote mime type is process specific
-MIME_KEEPNOTE = "application/x-keepnote" + str(random.randint(1, 100000))
+# richtext mime type is process specific
+MIME_RICHTEXT = "application/x-richtext" + str(random.randint(1, 100000))
 MIME_IMAGES = ["image/png",
                "image/bmp",
                "image/jpeg",
@@ -120,6 +124,9 @@ MIME_TEXT = ["text/plain",
              "COMPOUND_TEXT",
              "TEXT"]
 
+MIME_HTML = ["HTML Format",
+             "text/html"]
+
 
 # globals
 _g_clipboard_contents = None
@@ -148,6 +155,7 @@ def parse_utf(text):
         len(text) > 3 and text[3] == '\x00'):
         return text.decode("utf16")
     else:
+        text = text.replace("\x00", "")
         return unicode(text, "utf8")
 
 
@@ -201,19 +209,26 @@ class RichTextIO (object):
         self._html_buffer = HtmlBuffer()
 
     
-    def save(self, textbuffer, filename, title=None):
-        """Save buffer contents to file"""
+    def save(self, textbuffer, filename, title=None, stream=None):
+        """
+        Save buffer contents to file
+
+        textbuffer -- richtextbuffer to save
+        filename   -- HTML filename to save to (optional if stream given)
+        title      -- title of HTML file (optional)
+        stream     -- output stream for HTML file (optional)
+        """
         
-        path = os.path.dirname(filename)
-        self._save_images(textbuffer, path)
+        self._save_images(textbuffer, filename)
         
         try:
-            buffer_contents = iter_buffer_contents(textbuffer,
-                                                   None,
-                                                   None,
-                                                   ignore_tag)
+            buffer_contents = iter_buffer_contents(
+                textbuffer, None, None, ignore_tag)
             
-            out = safefile.open(filename, "wb", codec="utf-8")
+            if stream:
+                out = stream
+            else:
+                out = safefile.open(filename, "wb", codec="utf-8")
             self._html_buffer.set_output(out)
             self._html_buffer.write(buffer_contents,
                                     textbuffer.tag_table,
@@ -225,14 +240,21 @@ class RichTextIO (object):
         textbuffer.set_modified(False)
     
     
-    def load(self, textview, textbuffer, filename):
-        """Load buffer with data from file"""
+    def load(self, textview, textbuffer, filename, stream=None):
+        """
+        Load buffer with data from file
+
+        textbuffer -- richtextbuffer to load
+        filename   -- HTML filename to load (optional if stream given)
+        stream     -- output stream for HTML file (optional)
+        """
         
         # unhook expensive callbacks
         textbuffer.block_signals()
-        spell = textview.is_spell_check_enabled()
-        textview.enable_spell_check(False)
-        textview.set_buffer(None)
+        if textview:
+            spell = textview.is_spell_check_enabled()
+            textview.enable_spell_check(False)
+            textview.set_buffer(None)
 
 
         # clear buffer        
@@ -240,39 +262,37 @@ class RichTextIO (object):
         
         err = None
         try:
-            #from rasmus import util
-            #util.tic("read")
-            buffer_contents = list(self._html_buffer.read(
-                safefile.open(filename, "r", codec="utf-8")))
-            #util.toc()
-            
-            #util.tic("read2")            
+            if stream:
+                infile = stream
+            else:
+                infile = safefile.open(filename, "r", codec="utf-8")
+            buffer_contents = self._html_buffer.read(infile)
             textbuffer.insert_contents(buffer_contents,
                                        textbuffer.get_start_iter())
-            #util.toc()
+            infile.close()
 
             # put cursor at begining
             textbuffer.place_cursor(textbuffer.get_start_iter())
             
         except (HtmlError, IOError, Exception), e:
             err = e
-            
-            # TODO: turn into function
             textbuffer.clear()
-            textview.set_buffer(textbuffer)
+            if textview:
+                textview.set_buffer(textbuffer)
             ret = False            
         else:
             # finish loading
-            path = os.path.dirname(filename)
-            self._load_images(textbuffer, path)
-            textview.set_buffer(textbuffer)
-            textview.show_all()
+            self._load_images(textbuffer, filename)
+            if textview:
+                textview.set_buffer(textbuffer)
+                textview.show_all()
             ret = True
         
         # rehook up callbacks
         textbuffer.unblock_signals()
-        textview.enable_spell_check(spell)
-        textview.enable()
+        if textview:
+            textview.enable_spell_check(spell)
+            textview.enable()
         
         textbuffer.set_modified(False)
         
@@ -282,40 +302,61 @@ class RichTextIO (object):
         
 
     
-    def _load_images(self, textbuffer, path):
+    def _load_images(self, textbuffer, html_filename):
         """Load images present in textbuffer"""
-
-        for kind, it, param in iter_buffer_contents(textbuffer,
-                                                    None, None,
-                                                    ignore_tag):
-            if kind == "anchor":
-                child, widgets = param
-                    
-                if isinstance(child, RichTextImage):
-                    filename = child.get_filename()
-                    if is_relative_file(filename):
-                        filename = os.path.join(path, filename)
-                    
-                    child.set_from_file(filename)
+        
+        for kind, it, param in iter_buffer_anchors(textbuffer, None, None):
+            child, widgets = param
+            if isinstance(child, RichTextImage):
+                self._load_image(textbuffer, child, html_filename)
 
     
-    def _save_images(self, textbuffer, path):
+    def _save_images(self, textbuffer, html_filename):
         """Save images present in text buffer"""
+
+        for kind, it, param in iter_buffer_anchors(textbuffer, None, None):
+            child, widgets = param
+            if isinstance(child, RichTextImage):
+                self._save_image(textbuffer, child, html_filename)
+
+
+    def _load_image(self, textbuffer, image, html_filename):
+        image.set_from_file(
+            self._get_filename(html_filename, image.get_filename()))
+
+    def _save_image(self, textbuffer, image, html_filename):
+        if child.save_needed():
+            image.write(self._get_filename(html_filename, image.get_filename()))
+
+    def _get_filename(self, html_filename, filename):
+        if is_relative_file(filename):
+            path = os.path.dirname(html_filename)
+            return os.path.join(path, filename)
+        return filename
+
+                    
+
+class RichTextDragDrop (object):
+    """Manages drag and drop events for a richtext editor"""
+
+    def __init__(self, targets=[]):
+        self._acceptable_targets = []
+        self._acceptable_targets.extend(targets)
+
+    def append_target(self, target):
+        self._acceptable_targets.append(target)
+
+    def extend_targets(self, targets):
+        self._acceptable_targets.extend(target)
+
+    def find_acceptable_target(self, targets):
         
-        for kind, it, param in iter_buffer_contents(textbuffer,
-                                                    None, None,
-                                                    ignore_tag):
-            if kind == "anchor":
-                child, widgets = param
-                    
-                if isinstance(child, RichTextImage):
-                    filename = child.get_filename()
-                    if is_relative_file(filename):
-                        filename = os.path.join(path, filename)
-                        
-                    if child.save_needed():
-                        child.write(filename)
-                    
+        for target in self._acceptable_targets:
+            if target in targets:
+                return target
+        return None
+        
+
 
 
 
@@ -330,6 +371,11 @@ class RichTextView (gtk.TextView):
         self._blank_buffer = RichTextBuffer()
         self._popup_menu = None
         self._html_buffer = HtmlBuffer()
+        self._accel_group = None
+        self._accel_path = CONTEXT_MENU_ACCEL_PATH
+        self.dragdrop = RichTextDragDrop(MIME_IMAGES +  ["text/uri-list"] +
+                                         MIME_HTML + MIME_TEXT)
+        
 
         if textbuffer is None:
             textbuffer = RichTextBuffer() 
@@ -365,6 +411,7 @@ class RichTextView (gtk.TextView):
 
         #self.connect("button-press-event", self.on_button_press)
         self.connect("populate-popup", self.on_popup)
+
         
         # popup menus
         self.init_menus()
@@ -441,6 +488,15 @@ class RichTextView (gtk.TextView):
             # add all deferred anchors
             self._textbuffer.add_deferred_anchors(self)
 
+
+    def set_accel_group(self, accel_group):
+        self._accel_group = accel_group
+
+
+    def set_accel_path(self, accel_path):
+        self._accel_path = accel_path
+
+
     #======================================================
     # keyboard callbacks
 
@@ -535,9 +591,9 @@ class RichTextView (gtk.TextView):
             return
 
         if it is None:            
-            it = self._textbuffer.get_iter_at_mark(self._textbuffer.get_insert())
+            it = self._textbuffer.get_insert_iter()
 
-        for tag in it.get_tags():
+        for tag in chain(it.get_tags(), it.get_toggled_tags(False)):
             if isinstance(tag, RichTextLinkTag):
                 self.emit("visit-url", tag.get_href())
                 return True
@@ -555,36 +611,10 @@ class RichTextView (gtk.TextView):
         if not self._textbuffer:
             return
 
-        # in preference order
-        accepted_targets = MIME_IMAGES + \
-                           ["text/uri-list",
-                            "text/html",
-                            "text/plain"]
+        target = self.dragdrop.find_acceptable_target(drag_context.targets)
+        if target:
+            textview.drag_dest_set_target_list([(target, 0, 0)])
 
-        for target in accepted_targets:
-            if target in drag_context.targets:
-                textview.drag_dest_set_target_list([(target, 0, 0)])
-                break
-
-        '''
-        # check for image targets
-        img_target = self.drag_dest_find_target(drag_context, 
-                                                [(x, 0, 0)
-                                                 for x in MIME_IMAGES])
-
-             
-        if img_target is not None and img_target != "NONE":
-            textview.drag_dest_set_target_list([(img_target, 0, 0)])
-            
-        elif "application/pdf" in drag_context.targets:
-            textview.drag_dest_set_target_list([("application/pdf", 0, 0)])
-
-        elif "text/html" in drag_context.targets:
-            textview.drag_dest_set_target_list([("text/html", 0, 0)])
-            
-        else:
-            textview.drag_dest_set_target_list([("text/plain", 0, 0)])
-        '''
     
     
     def on_drag_data_received(self, widget, drag_context, x, y,
@@ -594,11 +624,11 @@ class RichTextView (gtk.TextView):
         if not self._textbuffer:
             return
         
-        img_target = self.drag_dest_find_target(drag_context, 
-                                                [(x, 0, 0)
-                                                 for x in MIME_IMAGES])        
-             
-        if img_target not in (None, "NONE"):
+        #TODO: make this pluggable.
+        
+        target = self.dragdrop.find_acceptable_target(drag_context.targets)
+        
+        if target in MIME_IMAGES:
             # process image drop
             pixbuf = selection_data.get_pixbuf()
             
@@ -611,8 +641,7 @@ class RichTextView (gtk.TextView):
                 drag_context.finish(True, True, eventtime)
                 self.stop_emission("drag-data-received")
 
-        elif self.drag_dest_find_target(drag_context, 
-            [("text/uri-list", 0, 0)]) not in (None, "NONE"):
+        elif target == "text/uri-list":
             # process URI drop
 
             uris = parse_utf(selection_data.data)
@@ -628,34 +657,22 @@ class RichTextView (gtk.TextView):
             self.insert_html("<br />".join(links))
             
                 
-        #elif self.drag_dest_find_target(drag_context, 
-        #    [("application/pdf", 0, 0)]) not in (None, "NONE"):
-        #    # process pdf drop
-        #    
-        #    data = selection_data.data
-        #    self.drop_pdf(data)
-        #    
-        #    drag_context.finish(True, True, eventtime)
-        #    self.stop_emission("drag-data-received")
-
-        elif self.drag_dest_find_target(drag_context, 
-            [("text/html", 0, 0)]) not in (None, "NONE"):
+        elif target in MIME_HTML:
             # process html drop
 
             html = parse_utf(selection_data.data)
-            #html = 
+            if taget == "HTML Format":
+                # skip over headers
+                html = html[html.find("\r\n\r\n")+4:]
+
             self.insert_html(html)
             
         
-        elif self.drag_dest_find_target(drag_context, 
-            [("text/plain", 0, 0)]) not in (None, "NONE"):
+        elif target in MIME_TEXT:
             # process text drop
-
-            #self._textbuffer.begin_user_action()
-            print "drop", repr(selection_data.get_text())
             self._textbuffer.insert_at_cursor(selection_data.get_text())
-            #self._textbuffer.end_user_action()
-            
+
+
 
     def on_drag_data_get(self, widget, drag_context, selection_data,
                              info, timestamp):
@@ -684,50 +701,6 @@ class RichTextView (gtk.TextView):
         '''
 
 
-
-    def drop_pdf(self, data):
-        """Drop a PDF into the TextView"""
-
-        if not self._textbuffer:
-            return
-
-        # NOTE: requires hardcoded convert
-        # TODO: generalize
-        
-        self._textbuffer.begin_user_action()
-        
-        try:
-            f, imgfile = tempfile.mkstemp(".png", "keepnote")
-            os.close(f)
-
-            out = os.popen("convert - %s" % imgfile, "wb")
-            out.write(data)
-            out.close()
-            
-            name, ext = os.path.splitext(imgfile)
-            imgfile2 = name + "-0" + ext
-            
-            if os.path.exists(imgfile2):
-                i = 0
-                while True:
-                    imgfile = name + "-" + str(i) + ext
-                    if not os.path.exists(imgfile):
-                        break
-                    self.insert_image_from_file(imgfile)
-                    os.remove(imgfile)
-                    i += 1
-                    
-            elif os.path.exists(imgfile):
-                
-                self.insert_image_from_file(imgfile)
-                os.remove(imgfile)
-        except:
-            if os.path.exists(imgfile):
-                os.remove(imgfile)
-
-        self._textbuffer.end_user_action()
-        
-
     
     #==================================================================
     # Copy and Paste
@@ -747,7 +720,8 @@ class RichTextView (gtk.TextView):
 
     
     def _on_paste(self):
-        """Callback for paste action"""    
+        """Callback for paste action"""
+        
         clipboard = self.get_clipboard(selection=CLIPBOARD_NAME)
         self.stop_emission('paste-clipboard')
         self.paste_clipboard(clipboard, None, self.get_editable())
@@ -756,6 +730,8 @@ class RichTextView (gtk.TextView):
     def copy_clipboard(self, clipboard):
         """Callback for copy event"""
 
+        #clipboard.set_can_store(None)
+
         if not self._textbuffer:
             return
     
@@ -773,9 +749,9 @@ class RichTextView (gtk.TextView):
            contents[0][0] == "anchor" and \
            isinstance(contents[0][2][0], RichTextImage):
             # copy image
-            targets = [(MIME_KEEPNOTE, gtk.TARGET_SAME_APP, RICHTEXT_ID),
-                       ("text/html", 0, RICHTEXT_ID)] + \
-                      [(x, 0, RICHTEXT_ID) for x in MIME_IMAGES]
+            targets = [(MIME_RICHTEXT, gtk.TARGET_SAME_APP, RICHTEXT_ID)] + \
+                [(x, 0, RICHTEXT_ID) for x in MIME_HTML] + \
+                [(x, 0, RICHTEXT_ID) for x in MIME_IMAGES]
             
             clipboard.set_with_data(targets, self._get_selection_data, 
                                     self._clear_selection_data,
@@ -783,14 +759,15 @@ class RichTextView (gtk.TextView):
 
         else:
             # copy text
-            targets = [(MIME_KEEPNOTE, gtk.TARGET_SAME_APP, RICHTEXT_ID),
-                       ("text/html", 0, RICHTEXT_ID)] + \
-                      [(x, 0, RICHTEXT_ID) for x in MIME_TEXT]
+            targets = [(MIME_RICHTEXT, gtk.TARGET_SAME_APP, RICHTEXT_ID)] + \
+                [(x, 0, RICHTEXT_ID) for x in MIME_HTML] + \
+                [(x, 0, RICHTEXT_ID) for x in MIME_TEXT]
             
             text = start.get_text(end)
             clipboard.set_with_data(targets, self._get_selection_data, 
                                     self._clear_selection_data,
-                                    (contents, text))        
+                                    (contents, text))
+
 
 
     def cut_clipboard(self, clipboard, default_editable):
@@ -814,7 +791,7 @@ class RichTextView (gtk.TextView):
             # nothing on clipboard
             return
         targets = set(targets)
-
+        
         
         # check that insert is allowed
         it = self._textbuffer.get_iter_at_mark(self._textbuffer.get_insert())
@@ -822,31 +799,34 @@ class RichTextView (gtk.TextView):
             return
 
         
-        if MIME_KEEPNOTE in targets:
-            # request KEEPNOTE contents object
-            clipboard.request_contents(MIME_KEEPNOTE, self._do_paste_object)
+        if MIME_RICHTEXT in targets:
+            # request RICHTEXT contents object
+            clipboard.request_contents(MIME_RICHTEXT, self._do_paste_object)
+            return
             
-        elif "text/html" in targets:
-            # request HTML
-            clipboard.request_contents("text/html", self._do_paste_html)
-            
-        else:
+        for mime_html in MIME_HTML:
+            if mime_html in targets:
+                # request HTML
+                if mime_html == "HTML Format":
+                    clipboard.request_contents(mime_html, 
+                                               self._do_paste_html_headers)
+                else:
+                    clipboard.request_contents(mime_html, self._do_paste_html)
+                return
 
-            # test image formats
-            for mime_image in MIME_IMAGES:
-                if mime_image in targets:
-                    clipboard.request_contents(mime_image,
-                                               self._do_paste_image)
-                    break
-            else:
-                # request text
-                clipboard.request_text(self._do_paste_text)
+        # test image formats
+        for mime_image in MIME_IMAGES:
+            if mime_image in targets:
+                clipboard.request_contents(mime_image, self._do_paste_image)
+                return
+
+        # request text
+        clipboard.request_text(self._do_paste_text)
 
     
     def paste_clipboard_as_text(self):
         """Callback for paste action"""    
         clipboard = self.get_clipboard(selection=CLIPBOARD_NAME)
-        #self.paste_clipboard(clipboard, None, self.get_editable())
 
         if not self._textbuffer:
             return
@@ -868,9 +848,12 @@ class RichTextView (gtk.TextView):
     def _do_paste_text(self, clipboard, text, data):
         """Paste text into buffer"""
         
+        if text is None:
+            return
+        
         self._textbuffer.begin_user_action()
         self._textbuffer.delete_selection(False, True)
-        self._textbuffer.insert_at_cursor(text)
+        self._textbuffer.insert_at_cursor(sanitize_text(text))
         self._textbuffer.end_user_action()
 
         self.scroll_mark_onscreen(self._textbuffer.get_insert())
@@ -878,14 +861,41 @@ class RichTextView (gtk.TextView):
     def _do_paste_html(self, clipboard, selection_data, data):
         """Paste HTML into buffer"""
 
-        # TODO: figure out right way to parse selection.data
         html = parse_utf(selection_data.data)        
-        self._textbuffer.begin_user_action()
-        self._textbuffer.delete_selection(False, True)
-        self.insert_html(html)
-        self._textbuffer.end_user_action()
         
-        self.scroll_mark_onscreen(self._textbuffer.get_insert())
+        try:
+            self._textbuffer.begin_user_action()
+            self._textbuffer.delete_selection(False, True)
+            self.insert_html(html)
+            self._textbuffer.end_user_action()
+        
+            self.scroll_mark_onscreen(self._textbuffer.get_insert())
+        except Exception, e:
+            pass
+
+
+    def _do_paste_html_headers(self, clipboard, selection_data, data):
+        """Paste HTML into buffer"""
+
+        html = parse_utf(selection_data.data)
+
+        # skip over headers
+        index = html.find("<!--StartFragment")
+        if index == -1:
+            return
+        index = html.find(">", index)
+        html = html[index+1:]
+        
+        try:
+            self._textbuffer.begin_user_action()
+            self._textbuffer.delete_selection(False, True)
+            self.insert_html(html)
+            self._textbuffer.end_user_action()
+        
+            self.scroll_mark_onscreen(self._textbuffer.get_insert())
+        except Exception, e:
+            pass
+    
     
     def _do_paste_image(self, clipboard, selection_data, data):
         """Paste image into buffer"""
@@ -918,6 +928,7 @@ class RichTextView (gtk.TextView):
     def _get_selection_data(self, clipboard, selection_data, info, data):
         """Callback for when Clipboard needs selection data"""
 
+        
         global _g_clipboard_contents
 
         contents, text = data
@@ -925,11 +936,12 @@ class RichTextView (gtk.TextView):
         _g_clipboard_contents = contents
 
         
-        if MIME_KEEPNOTE in selection_data.target:
+        if MIME_RICHTEXT in selection_data.target:
             # set rich text
-            selection_data.set(MIME_KEEPNOTE, 8, "<keepnote>")
+            selection_data.set(MIME_RICHTEXT, 8, "<richtext>")
             
-        elif "text/html" in selection_data.target:
+        elif len([x for x in MIME_HTML
+                  if x in selection_data.target]) > 0:
             # set html
             stream = StringIO.StringIO()
             self._html_buffer.set_output(stream)
@@ -1009,40 +1021,25 @@ class RichTextView (gtk.TextView):
     
     def on_popup(self, textview, menu):
         """Popup menu for RichTextView"""
-
-        self._popup_menu = menu
+        
+        self._popup_menu = menu        
 
         # insert "paste as plain text" after paste
         item = gtk.ImageMenuItem(stock_id=gtk.STOCK_PASTE,
                                  accel_group=None)        
         item.child.set_text(_("Paste As Plain Text"))
         item.connect("activate", lambda item: self.paste_clipboard_as_text())
+        #item.add_accelerator("activate", self._accel_group, ord("V"),
+        #                     gtk.gdk.CONTROL_MASK | gtk.gdk.SHIFT_MASK,
+        #                     gtk.ACCEL_VISIBLE)
         item.show()
         menu.insert(item, 3)
 
-    '''
-        menu.foreach(lambda item: menu.remove(item))
 
-        # Create the menu item
-        copy_item = gtk.MenuItem("Copy")
-        copy_item.connect("activate", self.on_copy)
-        menu.add(copy_item)
-        
-        accel_group = menu.get_accel_group()
-        print "accel", accel_group
-        if accel_group == None:
-            accel_group = gtk.AccelGroup()
-            menu.set_accel_group(accel_group)
-            print "get", menu.get_accel_group()
+        menu.set_accel_path(self._accel_path)
+        if self._accel_group:
+            menu.set_accel_group(self._accel_group)
 
-
-        # Now add the accelerator to the menu item. Note that since we created
-        # the menu item with a label the AccelLabel is automatically setup to 
-        # display the accelerators.
-        copy_item.add_accelerator("activate", accel_group, ord("C"),
-                                  gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
-        copy_item.show()
-    '''
     
 
     def _on_child_popup_menu(self, textbuffer, child, button, activate_time):
@@ -1123,7 +1120,7 @@ class RichTextView (gtk.TextView):
         if not self._textbuffer:
             return
         
-        contents = list(self._html_buffer.read([html],
+        contents = list(self._html_buffer.read(StringIO.StringIO(html),
                                                partial=True,
                                                ignore_errors=True))
 
@@ -1133,10 +1130,14 @@ class RichTextView (gtk.TextView):
             # download images included in html
             if kind == "anchor" and isinstance(param[0], RichTextImage):
                 img = param[0]
-                if img.get_filename().startswith("http:") or \
-                   img.get_filename().startswith("file:"):
-                    # TODO: protect this with exceptions
-                    img.set_from_url(img.get_filename(), "image.png")
+                filename = img.get_filename()
+                if filename and (filename.startswith("http:") or 
+                                 filename.startswith("file:")):
+                    try:
+                        img.set_from_url(filename, "image.png")
+                    except:
+                        # Be robust to errors from loading from the web.
+                        pass
         
         # add to buffer
         self._textbuffer.insert_contents(contents)
@@ -1164,8 +1165,6 @@ class RichTextView (gtk.TextView):
 
     #==========================================================
     # Find/Replace
-
-    # TODO: add wrapping to search
     
     def forward_search(self, it, text, case_sensitive, wrap=True):
         """Finds next occurrence of 'text' searching forwards"""
@@ -1335,9 +1334,9 @@ class RichTextView (gtk.TextView):
         if not tag:
             tag = self._textbuffer.tag_table.lookup(
                 RichTextLinkTag.tag_name(""))
-        
         self._textbuffer.toggle_tag_selected(tag)
 
+
     
     def set_font_family(self, family):
         """Sets the family font of the selection"""
@@ -1473,3 +1472,49 @@ gobject.signal_new("child-activated", RichTextView, gobject.SIGNAL_RUN_LAST,
 gobject.signal_new("visit-url", RichTextView, gobject.SIGNAL_RUN_LAST, 
     gobject.TYPE_NONE, (str,))
 
+
+
+
+'''
+    def drop_pdf(self, data):
+        """Drop a PDF into the TextView"""
+
+        if not self._textbuffer:
+            return
+
+        # NOTE: requires hardcoded convert
+        # TODO: generalize
+        
+        self._textbuffer.begin_user_action()
+        
+        try:
+            f, imgfile = tempfile.mkstemp(".png", "pdf")
+            os.close(f)
+
+            out = os.popen("convert - %s" % imgfile, "wb")
+            out.write(data)
+            out.close()
+            
+            name, ext = os.path.splitext(imgfile)
+            imgfile2 = name + "-0" + ext
+            
+            if os.path.exists(imgfile2):
+                i = 0
+                while True:
+                    imgfile = name + "-" + str(i) + ext
+                    if not os.path.exists(imgfile):
+                        break
+                    self.insert_image_from_file(imgfile)
+                    os.remove(imgfile)
+                    i += 1
+                    
+            elif os.path.exists(imgfile):
+                
+                self.insert_image_from_file(imgfile)
+                os.remove(imgfile)
+        except:
+            if os.path.exists(imgfile):
+                os.remove(imgfile)
+
+        self._textbuffer.end_user_action()
+    '''
