@@ -26,17 +26,21 @@
 
 
 # python imports
+import sys
 import re
 from HTMLParser import HTMLParser
 from xml.sax.saxutils import escape
 import urllib
 
 # keepnote imports
-from keepnote.gui.richtext.textbuffer_tools import \
+from keepnote import log_error, log_message
+
+from .textbuffer_tools import \
      iter_buffer_contents, \
      buffer_contents_iter_to_offset, \
      normalize_tags, \
      insert_buffer_contents, \
+     sanitize_text, \
      buffer_contents_apply_tags, \
      TextBufferDom, \
      TextDom, \
@@ -45,12 +49,12 @@ from keepnote.gui.richtext.textbuffer_tools import \
      TagNameDom
 
 
-from keepnote.gui.richtext.richtextbuffer import \
+from .richtextbuffer import \
      RichTextBuffer, \
      RichTextImage, \
      RichTextHorizontalRule
 
-from keepnote.gui.richtext.richtext_tags import \
+from .richtext_tags import \
      RichTextTag, \
      RichTextModTag, \
      RichTextFamilyTag, \
@@ -224,7 +228,7 @@ def unnest_indent_tags(contents):
                 yield item
 
         else:
-                yield item
+            yield item
                 
 
 def find_paragraphs(contents):
@@ -313,53 +317,61 @@ def parse_css_style(stylestr):
     for statement in stylestr.split(";"):
         statement = statement.strip()            
 
-        if statement.startswith("font-size"):
-            # font size
-            size = int("".join(filter(lambda x: x.isdigit(),
-                               statement.split(":")[1])))
-            yield "size " + str(size)
+        try:
 
-        elif statement.startswith("font-family"):
-            # font family
-            yield "family " + statement.split(":")[1].strip()
+            if statement.startswith("font-size"):
+                # font size
+                size = int(float(
+                        "".join(filter(lambda x: x.isdigit() or x == ".",
+                                          statement.split(":")[1]))))
+                yield "size " + str(size)
 
-        elif statement.startswith("text-align"):
-            # text justification
-            align = statement.split(":")[1].strip()
+            elif statement.startswith("font-family"):
+                # font family
+                yield "family " + statement.split(":")[1].strip()
 
-            if align not in JUSTIFY_VALUES:
-                raise HtmlError("unknown justification '%s'" % align)
+            elif statement.startswith("text-align"):
+                # text justification
+                align = statement.split(":")[1].strip()
 
-            if align == "justify":
-                yield "fill"
-            else:
-                yield align
+                #if align not in JUSTIFY_VALUES:
+                #    raise HtmlError("unknown justification '%s'" % align)
 
-        elif statement.startswith("color"):
-            # foreground color
-            fg_color = statement.split(":")[1].strip()
+                if align == "justify":
+                    yield "fill"
+                else:
+                    yield align
 
-            if fg_color.startswith("#"):
-                if len(fg_color) == 4:
-                    x, a, b, c = fg_color
-                    fg_color = x + a + a + b + b+ c + c
+            elif statement.startswith("color"):
+                # foreground color
+                fg_color = statement.split(":")[1].strip()
 
-                if len(fg_color) == 7:
-                    yield "fg_color " + fg_color
+                if fg_color.startswith("#"):
+                    if len(fg_color) == 4:
+                        x, a, b, c = fg_color
+                        fg_color = x + a + a + b + b+ c + c
+
+                    if len(fg_color) == 7:
+                        yield "fg_color " + fg_color
 
 
-        elif statement.startswith("background-color"):
-            # background color
-            bg_color = statement.split(":")[1].strip()
+            elif statement.startswith("background-color"):
+                # background color
+                bg_color = statement.split(":")[1].strip()
 
-            if bg_color.startswith("#"):
-                if len(bg_color) == 4:
-                    x, a, b, c = bg_color
-                    bg_color = x + a + a + b + b+ c + c
+                if bg_color.startswith("#"):
+                    if len(bg_color) == 4:
+                        x, a, b, c = bg_color
+                        bg_color = x + a + a + b + b + c + c
 
-                if len(bg_color) == 7:
+                    if len(bg_color) == 7:
                         yield "bg_color " + bg_color
 
+        except:
+            # ignore css statements that we cannot parse
+            pass
+
+
         
 
 class HtmlTagDom (TagDom):
@@ -430,7 +442,9 @@ class HtmlTagModReader (HtmlTagReader):
 
     html2buffer_tag = {
         "b": "bold",
+        "strong": "bold",
         "i": "italic",
+        "em": "italic",
         "u": "underline",
         "strike": "strike",
         "tt": "tt",
@@ -504,6 +518,41 @@ class HtmlTagSpanReader (HtmlTagReader):
                     self._io.append_child(TagNameDom(tagstr), True)
 
 
+class HtmlTagFontReader (HtmlTagReader):
+    """<font> tags"""
+
+    def __init__(self, io):
+        HtmlTagReader.__init__(self, io, "font")
+
+
+    def parse_starttag(self, htmltag, attrs):
+        for key, value in attrs:
+            if key == "style":
+                for tagstr in parse_css_style(value):
+                    self._io.append_child(TagNameDom(tagstr), True)
+            elif key == "face":
+                self._io.append_child(TagNameDom("family " + value), True)
+            elif key == "size":
+                size = int(value)
+                if size < 0:
+                    sizept = 4
+                else:
+                    sizept = 4 + size * 2 #(default should be 10pt)
+                
+                self._io.append_child(TagNameDom("size " + str(sizept)), True)
+
+
+class HtmlTagCenterReader (HtmlTagReader):
+    """<center> tags"""
+
+    def __init__(self, io):
+        HtmlTagReader.__init__(self, io, "center")
+
+
+    def parse_starttag(self, htmltag, attrs):
+        self._io.append_child(TagNameDom("center"), True)
+
+
 class HtmlTagDivReader (HtmlTagReader):
 
     def __init__(self, io):
@@ -597,7 +646,25 @@ class HtmlTagParReader (HtmlTagReader):
         HtmlTagReader.__init__(self, io, "p")
 
     def parse_starttag(self, htmltag, attrs):
-        self._io.append_text("\n")
+
+        # do not create a newline if one already exists
+        last_child = self._io.last_child()
+        #print last_child
+        #if isinstance(last_child, TextDom):
+        #    print last_child.lst
+
+        if (not isinstance(last_child, TextDom) or 
+            not last_child.get().endswith("\n")):
+            self._io.append_text("\n")
+
+
+        for key, value in attrs:
+            if key == "style":
+                for tagstr in parse_css_style(value):
+                    self._io.append_child(TagNameDom(tagstr), True)
+            elif key == "align":
+                if value.lower() in ("left", "right", "center"):
+                    self._io.append_child(TagNameDom(value.lower()), True)
 
     def parse_endtag(self, htmltag):
         self._io.append_text("\n")
@@ -614,14 +681,16 @@ class HtmlTagListItemReader (HtmlTagReader):
         for key, value in attrs:
             if key == "style":
                 for statement in value.split(";"):
-                    key2, value2 = statement.split(":")
-                    value2 = value2.strip()
+                    tokens = statement.split(":")
+                    if len(tokens) == 2:
+                        key2, value2 = tokens
+                        value2 = value2.strip()
 
-                    if key2.strip() == "list-style-type":
-                        if value2 == "disc":
-                            par_type = "bullet"
-                        elif value2 == "none":
-                            par_type = "none"
+                        if key2.strip() == "list-style-type":
+                            if value2 == "disc":
+                                par_type = "bullet"
+                            elif value2 == "none":
+                                par_type = "none"
 
         tag = TagNameDom("li %s" % par_type)
         self._io.append_child(tag, True)
@@ -667,11 +736,9 @@ class HtmlTagUnorderedListWriter (HtmlTagWriter):
         
     def write_tag_begin(self, out, dom, xhtml):
         out.write("<ul>")
-        #out.write("<ol>")
 
     def write_tag_end(self, out, dom, xhtml):
         out.write("</ul>\n")
-        #out.write("</ol>\n")
 
 class HtmlTagBulletWriter (HtmlTagWriter):
     
@@ -816,7 +883,9 @@ class HtmlBuffer (HTMLParser):
 
         # mod tags
         self.add_tag_reader(HtmlTagModReader(self, "b"))
+        self.add_tag_reader(HtmlTagModReader(self, "strong"))
         self.add_tag_reader(HtmlTagModReader(self, "i"))
+        self.add_tag_reader(HtmlTagModReader(self, "em"))
         self.add_tag_reader(HtmlTagModReader(self, "u"))
         self.add_tag_reader(HtmlTagModReader(self, "strike"))
         self.add_tag_reader(HtmlTagModReader(self, "tt"))
@@ -825,6 +894,8 @@ class HtmlBuffer (HTMLParser):
 
         # span/div readers
         self.add_tag_reader(HtmlTagSpanReader(self))
+        self.add_tag_reader(HtmlTagFontReader(self))
+        self.add_tag_reader(HtmlTagCenterReader(self))
         self.add_tag_reader(HtmlTagDivReader(self))
 
         # span/div writers
@@ -870,12 +941,16 @@ class HtmlBuffer (HTMLParser):
         self._tag_stack = [(None, self._dom)]
 
         try:
-            for line in infile:
-                self.feed(line)                
+            self.reset()
+            #for line in infile:
+            #    self.feed(line)
+            self.feed(infile.read())
             self.close()
             
         except Exception, e:
+            log_error(e, sys.exc_info()[2])
             # reraise error if not ignored
+            self.close()
             if not ignore_errors:
                 raise
         
@@ -883,6 +958,7 @@ class HtmlBuffer (HTMLParser):
         return unnest_indent_tags(self._dom.get_contents())
 
 
+
     def process_dom_read(self, dom):
         """Process a DOM after reading"""
 
@@ -893,7 +969,7 @@ class HtmlBuffer (HTMLParser):
                 if node.prev_sibling():
                     node.get_parent().insert_before(node, TextDom("\n"))
             
-            if isinstance(node, TagNameDom) and node.tagname.startswith("li "):
+            elif isinstance(node, TagNameDom) and node.tagname.startswith("li "):
                 # list items end with an implied newline
 
                 if not (isinstance(node.last_child(), TagNameDom) and \
@@ -902,15 +978,28 @@ class HtmlBuffer (HTMLParser):
             
             for child in list(node):
                 walk(child)
+
+            
+
         walk(dom)
 
+
+    def ignore_input(self):
+        """Returns True if currently ignoring input"""
+        
+        for tag, ptr in self._tag_stack:
+            if tag in ("style", "script"):
+                return True
+
+        return (not self._within_body and not self._partial)
+
     
     def append_text(self, text):
         
         if len(text) > 0:
             last_child = self._dom_ptr.last_child()
             if isinstance(last_child, TextDom):
-                last_child.text += text
+                last_child.append(text)
             else:
                 self._dom_ptr.append_child(TextDom(text))
 
@@ -918,43 +1007,47 @@ class HtmlBuffer (HTMLParser):
         self._dom_ptr.append_child(child)
         if visit:
             self._dom_ptr = child
+
+    def last_child(self):
+        return self._dom_ptr.last_child()
         
     
     def handle_starttag(self, htmltag, attrs):
         """Callback for parsing a starting HTML tag"""
-
+        
         self._newline = False
-
+        
         # start a new tag on htmltag stack
         self._tag_stack.append((htmltag, self._dom_ptr))
 
-
-        if htmltag == "html":
-            # ignore html tag
-            pass
-        
-        elif htmltag == "body":
-            # note that we are within the body tag
-            self._within_body = True
-
-        elif htmltag == "br":
+        if htmltag == "br":
             # insert newline
             self.append_text("\n")
             self._newline = True
         
         elif htmltag in self._tag_readers:
             # use tag parser
-            self._tag_readers[htmltag].parse_starttag(htmltag, attrs)
+            if not self.ignore_input():
+                self._tag_readers[htmltag].parse_starttag(htmltag, attrs)
+
+        elif htmltag == "html":
+            # ignore html tag
+            pass
+        
+        elif htmltag == "body":
+            # note that we are within the body tag
+            self._within_body = True
             
         else:
             # ingore other html tags
             pass
-        
+
+
         
 
     def handle_endtag(self, htmltag):
         """Callback for parsing a ending HTML tag"""
-
+        
         if not self._partial:
             if not self._within_body:
                 return
@@ -971,7 +1064,8 @@ class HtmlBuffer (HTMLParser):
 
         elif htmltag in self._tag_readers:
             # use tag parser
-            self._tag_readers[htmltag].parse_endtag(htmltag)
+            if not self.ignore_input():
+                self._tag_readers[htmltag].parse_endtag(htmltag)
 
 
         # pop dom stack
@@ -986,20 +1080,29 @@ class HtmlBuffer (HTMLParser):
     
     def handle_data(self, data):
         """Callback for character data"""
-
-        if not self._partial and not self._within_body:
+        
+        if self.ignore_input():
             return
-
+        
         if self._newline:
-            data = re.sub("^\n[\n ]*", "", data)
-            data = re.sub("[\n ]+", " ", data)
+            #data = re.sub("^\r?\n[\r\n\t ]*", "", data)
+            data = re.sub(self.remove_first_whitespace, "", data)
+
+            # collapse a sequence of whitespace into one space char
+            #data = re.sub("[\r\n\t ]+", " ", data)
+            data = re.sub(self.remove_whitespace, " ", data)
             self._newline = False
         else:
-            data = re.sub("[\n ]+", " ", data)
+            # collapse a sequence of whitespace into one space char
+            #data = re.sub("[\r\n\t ]+", " ", data)
+            data = re.sub(self.remove_whitespace, " ", data)
         
         if len(data) > 0:
             self.append_text(data)
 
+    remove_first_whitespace = re.compile("^\r?\n[\r\t\n ]*")
+    remove_whitespace = re.compile("[\r\n\t ]+")
+
     
     def handle_entityref(self, name):
         """Callback for reading entityref"""
@@ -1060,7 +1163,7 @@ class HtmlBuffer (HTMLParser):
         """Write DOM"""
         for child in dom:
             if isinstance(child, TextDom):
-                self.write_text(child.text, xhtml=xhtml)
+                self.write_text(child.get(), xhtml=xhtml)
 
             elif isinstance(child, TagDom):                
                 self.write_tag_begin(child, xhtml=xhtml)
@@ -1083,6 +1186,13 @@ class HtmlBuffer (HTMLParser):
         text = text.replace(">", "&gt;")
         text = text.replace("<", "&lt;")
         text = text.replace("\t", "&#09;")
+        text = text.replace("\n ", "\n&nbsp;")
+        text = text.replace("\r ", "\r&nbsp;")
+
+        if text.startswith(" "):
+            # protect leading space
+            text = "&nbsp;" + text[1:]
+
         text = text.replace("  ", " &nbsp;")
         if xhtml:
             text = text.replace("\n", "<br/>\n")
@@ -1100,8 +1210,7 @@ class HtmlBuffer (HTMLParser):
                 return
         
         # warning
-        #TODO:
-        print "unknown anchor element", anchor
+        log_message("unknown anchor element", anchor)
 
     
     def write_tag_begin(self, dom, xhtml=True):
@@ -1202,15 +1311,15 @@ class HtmlBuffer (HTMLParser):
             if isinstance(node, AnchorDom) and \
                isinstance(node.anchor, RichTextHorizontalRule) and \
                isinstance(last_leaf[0], TextDom) and \
-               last_leaf[0].text.endswith("\n"):
-                last_leaf[0].text = last_leaf[0].text[:-1]
+               last_leaf[0].get().endswith("\n"):
+                last_leaf[0].set(last_leaf[0].get()[:-1])
 
             # delete preceding newline of <ol> <ul>
             if isinstance(node, TagDom) and \
                isinstance(node.tag, RichTextIndentTag) and \
                isinstance(last_leaf[0], TextDom) and \
-               last_leaf[0].text.endswith("\n"):
-                last_leaf[0].text = last_leaf[0].text[:-1]
+               last_leaf[0].get().endswith("\n"):
+                last_leaf[0].set(last_leaf[0].get()[:-1])
                 
             # delete preceding newline of </li>
             if isinstance(node, LiHtmlTagDom):
@@ -1226,8 +1335,8 @@ class HtmlBuffer (HTMLParser):
                         child = child.last_child()
                 
                 if isinstance(child, TextDom) and \
-                   child.text.endswith("\n"):
-                    child.text = child.text[:-1]
+                   child.get().endswith("\n"):
+                    child.set(child.get()[:-1])
             
             if node.is_leaf():
                 # process leaves
@@ -1236,8 +1345,8 @@ class HtmlBuffer (HTMLParser):
                 if isinstance(last_leaf[0], AnchorDom) and \
                    isinstance(last_leaf[0].anchor, RichTextHorizontalRule) and \
                    isinstance(node, TextDom) and \
-                   node.text.startswith("\n"):
-                    node.text = node.text[1:]
+                   node.get().startswith("\n"):
+                    node.set(node.get()[1:])
 
                 # empty tags are skiped as leaves
                 if not isinstance(node, TagDom):
