@@ -37,32 +37,34 @@ pygtk.require('2.0')
 import gtk, gobject, pango
 from gtk import gdk
 
+# TODO: remove
 # keepnote imports
 import keepnote
 
 
 # textbuffer imports
-from keepnote.gui.richtext.textbuffer_tools import \
+from .textbuffer_tools import \
     move_to_start_of_line, \
     move_to_end_of_line, \
     iter_buffer_contents, \
+    iter_buffer_anchors, \
     buffer_contents_iter_to_offset, \
     normalize_tags, \
     insert_buffer_contents, \
     buffer_contents_apply_tags
 
 # richtext imports
-from keepnote.gui.richtext.richtextbasebuffer import \
+from .richtextbasebuffer import \
     RichTextBaseBuffer, \
     add_child_to_buffer, \
     RichTextAnchor
-from keepnote.gui.richtext.indent_handler import IndentHandler
-from keepnote.gui.richtext.font_handler import \
+from .indent_handler import IndentHandler
+from .font_handler import \
     FontHandler, RichTextBaseFont
 
 
 # richtext tags imports
-from keepnote.gui.richtext.richtext_tags import \
+from .richtext_tags import \
     RichTextTagTable, \
     RichTextTag, \
     RichTextModTag, \
@@ -256,8 +258,15 @@ class BaseImage (BaseWidget):
         self._img.set_from_stock(stock, size)
     
 
-# TODO: think about how a single anchor could manage multiple widgets in
-# multiple textviews
+
+def get_image_format(filename):
+    """Returns the image format for a filename"""
+    f, ext = os.path.splitext(filename)
+    ext = ext.replace(u".", "").lower()
+    if ext == "jpg":
+        ext = "jpeg"
+    return ext
+
 
 class RichTextImage (RichTextAnchor):
     """An Image child widget in a RichTextView"""
@@ -317,15 +326,25 @@ class RichTextImage (RichTextAnchor):
     
     def write(self, filename):
         """Write image to file"""
-
-        # TODO: move this to external function
-        f, ext = os.path.splitext(filename)
-        ext = ext.replace(u".", "").lower()
-        if ext == "jpg":
-            ext = "jpeg"
-
+        
         # TODO: make more checks on saving
-        self._pixbuf_original.save(filename, ext)
+        if self._pixbuf:
+            ext = get_image_format(filename)
+            self._pixbuf_original.save(filename, ext)
+            self._save_needed = False
+        
+
+    def write_stream(self, stream, filename="image.png"):
+        """
+        Write image to stream
+        'filename' is used to infer picture format only.
+        """
+
+        def write(buf):
+            stream.write(buf)
+            return True
+        format = get_image_format(filename)
+        self._pixbuf_original.save_to_callback(write, format)
         self._save_needed = False
         
         
@@ -337,9 +356,6 @@ class RichTextImage (RichTextAnchor):
         img._size = list(self.get_size())
         
         if self._pixbuf:
-            # TODO: this loop will not be necessary when add_view() is complete
-            #for widget in img.get_all_widgets().itervalues():
-            #    widget.set_from_pixbuf(self._pixbuf)
             img._pixbuf = self._pixbuf
             img._pixbuf_original = self._pixbuf_original
         else:
@@ -363,7 +379,7 @@ class RichTextImage (RichTextAnchor):
         try:
             self._pixbuf_original = gdk.pixbuf_new_from_file(filename)
             
-        except gobject.GError, e:
+        except Exception:
             # use missing image instead
             self.set_no_image()
         else:
@@ -376,6 +392,25 @@ class RichTextImage (RichTextAnchor):
             for widget in self.get_all_widgets().itervalues():
                 widget.set_from_pixbuf(self._pixbuf)
 
+    def set_from_stream(self, stream):
+
+        loader = gtk.gdk.PixbufLoader()
+        try:
+            loader.write(stream.read())
+            loader.close()
+            self._pixbuf_original = loader.get_pixbuf()
+        except Exception:
+            self.set_no_image()
+        else:
+            # successful image load, set its size
+            self._pixbuf = self._pixbuf_original
+
+            if self.is_size_set():
+                self.scale(self._size[0], self._size[1], False)
+
+            for widget in self.get_all_widgets().itervalues():
+                widget.set_from_pixbuf(self._pixbuf)
+
 
     def set_no_image(self):
         """Set the 'no image' icon"""
@@ -401,19 +436,22 @@ class RichTextImage (RichTextAnchor):
 
 
     
-    def set_from_url(self, url, filename):
+    def set_from_url(self, url, filename=None):
         """Set image by url"""
 
         imgfile = None        
 
         try:
             # make local temp file
-            f, imgfile = tempfile.mkstemp("", "keepnote")
+            f, imgfile = tempfile.mkstemp("", "image")
             os.close(f)
 
             if download_file(url, imgfile):
                 self.set_from_file(imgfile)
-                self.set_filename(filename)
+                if filename is not None:
+                    self.set_filename(filename)
+                else:
+                    self.set_filename(url)
             else:
                 raise Exception("Could not download file")
         except Exception:
@@ -605,6 +643,7 @@ class RichTextFont (RichTextBaseFont):
                 self.fg_color = tag.get_color()
             elif isinstance(tag, RichTextBGColorTag):
                 self.bg_color = tag.get_color()
+
                 
         # set indentation info
         for tag in chain(tags, current_tags):
@@ -614,6 +653,8 @@ class RichTextFont (RichTextBaseFont):
 
             elif isinstance(tag, RichTextLinkTag):
                 self.link = tag
+
+            
         
         
 
@@ -651,6 +692,7 @@ class RichTextBuffer (RichTextBaseBuffer):
         
         # set of all anchors in buffer
         self._anchors = set()
+        self._anchors_highlighted = set()
         #self._child_uninit = set()
 
         # anchors that still need to be added,
@@ -663,6 +705,7 @@ class RichTextBuffer (RichTextBaseBuffer):
         
         RichTextBaseBuffer.clear(self)
         self._anchors.clear()
+        self._anchors_highlighted.clear()
         self._anchors_deferred.clear()
 
 
@@ -670,7 +713,7 @@ class RichTextBuffer (RichTextBaseBuffer):
         """Inserts a content stream into the TextBuffer at iter 'it'"""
 
         if it is None:
-            it = self.get_iter_at_mark(self.get_insert())
+            it = self.get_insert_iter()
 
         self.begin_user_action()        
         insert_buffer_contents(self, it,
@@ -738,14 +781,18 @@ class RichTextBuffer (RichTextBaseBuffer):
         # let indent manager prepare the delete
         #if self.is_interactive():
         #    self._indent.prepare_delete_range(start, end)
-                
+        
         # call super class
         RichTextBaseBuffer._on_delete_range(self, textbuffer, start, end)
         
         # deregister any deleted anchors
-        for kind, offset, param in self._next_action.contents:
+        for kind, offset, param in iter_buffer_contents(
+            self, start, end, ignore_tag):
             if kind == "anchor":
-                self._anchors.remove(param[0])
+                child = param[0]
+                self._anchors.remove(child)
+                if child in self._anchors_highlighted:
+                    self._anchors_highlighted.remove(child)
 
     #=========================================
     # indentation interface
@@ -849,8 +896,7 @@ class RichTextBuffer (RichTextBaseBuffer):
 
 
     def _add_child_at_anchor(self, child, textview):
-        #print "add", child.get_widget()
-
+        
         # skip children whose insertion was rejected
         if child.get_deleted():
             return
@@ -860,20 +906,18 @@ class RichTextBuffer (RichTextBaseBuffer):
         textview.add_child_at_anchor(widget, child)
 
         child.show()
-        #widget.show_all()
-        #print "added", child.get_widget()
 
     
     def insert_image(self, image, filename="image.png"):
         """Inserts an image into the textbuffer at current position"""
-
+        
         # set default filename
         if image.get_filename() is None:
             image.set_filename(filename)
         
         # insert image into buffer
         self.begin_user_action()
-        it = self.get_iter_at_mark(self.get_insert())
+        it = self.get_insert_iter()
         self.add_child(it, image)
         image.show()
         self.end_user_action()
@@ -883,7 +927,7 @@ class RichTextBuffer (RichTextBaseBuffer):
         """Insert Horizontal Rule"""
         self.begin_user_action()
 
-        it = self.get_iter_at_mark(self.get_insert())
+        it = self.get_insert_iter()
         hr = RichTextHorizontalRule()
         self.add_child(it, hr)
         
@@ -957,9 +1001,9 @@ class RichTextBuffer (RichTextBaseBuffer):
             if len(sel) > 0:
                 it = sel[0]
             else:
-                it = self.get_iter_at_mark(self.get_insert())
+                it = self.get_insert_iter()
 
-        for tag in it.get_tags():
+        for tag in chain(it.get_tags(), it.get_toggled_tags(False)):
             if isinstance(tag, RichTextLinkTag):
                 start, end = self.get_tag_region(it, tag)
                 return tag, start, end
@@ -1011,41 +1055,25 @@ class RichTextBuffer (RichTextBaseBuffer):
     
     def highlight_children(self):
         """Highlight any children that are within selection range"""
-
-        # TODO: I once had an exception that said an anchor in self._anchors
-        # was already deleted.  I do not know yet how this got out of
-        # sync, seeing as I listen for deletes.
         
         sel = self.get_selection_bounds()
         focus = None
-
-        # TODO: speed this up.  Don't process all children
         
         if len(sel) > 0:
-            
-            # selection exists, get range (a, b)
-            a = sel[0].get_offset()
-            b = sel[1].get_offset()
-            for child in self._anchors:
-                it = self.get_iter_at_child_anchor(child)
-                offset = it.get_offset()
-                if a <= offset < b:
-                    child.highlight()
-                else:
+            highlight = set(x[2][0] for x in 
+                            iter_buffer_anchors(self, sel[0], sel[1]))
+            for child in self._anchors_highlighted:
+                if child not in highlight:
                     child.unhighlight()
-
-                for w in child.get_all_widgets().itervalues():
-                    top = w.get_toplevel()
-                    if top and isinstance(top, gtk.Window):
-                        f = top.get_focus()
-                        if f:
-                            focus = f
-            if focus:
-                focus.grab_focus()
+            for child in highlight:
+                child.highlight()
+            self._anchors_highlighted = highlight
+            
         else:
             # no selection, unselect all children
-            for child in self._anchors:
+            for child in self._anchors_highlighted:
                 child.unhighlight()
+            self._anchors_highlighted.clear()
 
 
 
