@@ -656,20 +656,21 @@ class ThreePaneViewer (Viewer):
                 return
             node = nodes[0]
         
-        treenodes = self.treeview.get_selected_nodes()
 
         if direct:
             # direct goto: open up treeview all the way to the node
             self.treeview.select_nodes([node])
         else:
-            # indirect goto: donot open up treeview, only listview
-            
+            # indirect goto: do not open up treeview, only listview
+        
+            treenodes = self.treeview.get_selected_nodes()
+    
             # get path to root
             path = []
             ptr = node
             while ptr:
                 if ptr in treenodes:
-                    # if parent path is allready selected then quit
+                    # if parent path is already selected then quit
                     path = []
                     break
                 path.append(ptr)
@@ -684,7 +685,10 @@ class ThreePaneViewer (Viewer):
             # make selections
             if node2:
                 self.treeview.select_nodes([node2])
-            self.listview.select_nodes([node])
+
+            # This test might be needed for windows crash
+            if node2 != node:
+                self.listview.select_nodes([node])
                     
 
 
